/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef __C4DBASESELECT_H
#define __C4DBASESELECT_H

#ifdef __API_INTERN__
	#include "baseselect.h"
#else

#include "operatingsystem.h"
#include "c4d_file.h"

struct BaseSelectData
{
	LONG a,b; 
};

class BaseSelect
{
	private:
		BaseSelect();
		~BaseSelect();
	public:
		LONG GetCount(void) const;
		LONG GetSegments(void) const;

		Bool Select(LONG num);           
		Bool SelectAll(LONG min, LONG max);

		Bool Deselect(LONG num);
		Bool DeselectAll(void);

		Bool Toggle(LONG num);
		Bool ToggleAll(LONG min, LONG max);

		Bool GetRange(LONG seg, LONG maxElements, LONG *a, LONG *b) const; // pass maxElements to make sure the returned values for a and b are < maxElements or MAXLONGl for no additional checks
		
		Bool IsSelected(LONG num) const { return C4DOS.Bs->IsSelected(this,num); }

		Bool CopyTo(BaseSelect *dest) const;
		Bool Merge(const BaseSelect *src);
		Bool Deselect(const BaseSelect *src);
		Bool Cross(const BaseSelect *src);
		BaseSelect *GetClone(void) const;

		Bool FromArray(UCHAR *selection, LONG count);
		UCHAR *ToArray(LONG count) const;

		void Write(HyperFile *hf);
		Bool Read(HyperFile *hf);

		static BaseSelect *Alloc(void);
		static void Free(BaseSelect *&bs);

		Bool FindSegment(LONG num, LONG *segment) const { return C4DOS.Bs->FindSegment(this,num,segment); }
		BaseSelectData *GetData()  { return C4DOS.Bs->GetData(this); }
		Bool CopyFrom(BaseSelectData *ndata, LONG ncnt) { return C4DOS.Bs->CopyFrom(this,ndata,ncnt); }

		LONG GetDirty() const { return C4DOS.Bs->GetDirty(this); }

		LONG GetLastElement(void) const { return C4DOS.Bs->GetLastElement(this); }
};

#endif

#endif
