/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

//---------------------------------------------------
#ifndef _C4D_BATCHRENDER_H_
#define _C4D_BATCHRENDER_H_

#ifdef __API_INTERN__
#include "c4d_basetime.h"
#include "ge_string.h"
#include "c4d_library.h"
#include "c4d_gui.h"
#else
#include "c4d.h"
#endif

class GeListNode;
class Filename;
class BaseBitmap;
class BaseDocument;
class String;
class BaseSound;

#define C4D_BATCHRENDER_LIBRARY_ID		465003507

#define BR_STOP		1
#define BR_START	2

//render flags
enum 
{
	RM_PROGRESS = 1000,
	RM_FINISHED,
	RM_STOPPED,
	RM_ERROR,
	RM_ERROR2,
	RM_PAUSED,
	RM_QUEUE, //means is in queue but not rendered yet

	RM_NONE,
};

class BatchRender
{
	private:
		BatchRender();
		~BatchRender();
	public:
		Bool Open(void);
		Bool AddFile(const Filename &File, LONG number);
		Bool DelFile(const Filename &File);

		Bool IsRendering();
		void SetRendering(LONG set);

		LONG GetElementCount();
		
		Filename GetElement(LONG n);
		void EnableElement(LONG n, Bool bSet);
		Bool GetEnableElement(LONG n);

		LONG GetElementStatus(LONG n);
};

BatchRender* GetBatchRender();

//---------------------------------------------------
//	---INTERNAL STUFF

class iBatchRender;

struct BatchRenderLibrary : public C4DLibrary
{
	iBatchRender*	(*GetBatchRender)();

	Bool					(iBatchRender::*Open)								();
	Bool					(iBatchRender::*AddFile)						(const Filename &File, LONG number);
	Bool					(iBatchRender::*DelFile)						(const Filename &File);

	Bool					(iBatchRender::*IsRendering)				();
	void					(iBatchRender::*SetRendering)				(LONG set);


	LONG					(iBatchRender::*GetElementCount)		();

	Filename			(iBatchRender::*GetElement)					(LONG n);
	void					(iBatchRender::*EnableElement)			(LONG n, Bool bSet);
	Bool					(iBatchRender::*GetEnableElement)		(LONG n);

	LONG					(iBatchRender::*GetElementStatus)		(LONG n);
};

#endif

