#ifndef __API_INTERN__

#include "lib_clipmap.h"
#include "c4d_basecontainer.h"

static GeClipMapLib *lib_GeClipMap = NULL;

static GeClipMapLib *CheckLib(LONG offset)
{
	return (GeClipMapLib*)CheckLib( LIB_CLIPMAP, offset, (C4DLibrary**)&lib_GeClipMap);
}

#define GeClipMapCall(b)      GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, b)); \
															if (!lib || !lib->b) return; \
															(((iGeClipMap*)this)->*lib->b)

#define GeClipMapCallR(a,b)   GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, b)); \
															if (!lib || !lib->b) return a; \
															return (((iGeClipMap*)this)->*lib->b)

GeClipMap* GeClipMap::Alloc()
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, Alloc));
	if (!lib || !lib->Alloc) return NULL;
	return lib->Alloc();
}

void GeClipMap::Free(GeClipMap *&data)
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, Free));
	if (!lib || !lib->Free) return;
	lib->Free( data );
}

GeClipMap::GeClipMap()
{
}

IMAGERESULT GeClipMap::Init(LONG w, LONG h, LONG bits)
{
	GeClipMapCallR( IMAGERESULT_NOTEXISTING, Init )( w, h, bits );
}

IMAGERESULT GeClipMap::Init(BaseBitmap *bm)
{
	GeClipMapCallR( IMAGERESULT_NOTEXISTING, InitWith )( bm );
}

IMAGERESULT GeClipMap::Init(BaseBitmap *bm, BaseBitmap *alpha_channel)
{
	GeClipMapCallR( IMAGERESULT_NOTEXISTING, InitFrom )( bm, alpha_channel );
}

IMAGERESULT GeClipMap::Init(const Filename &name, LONG frame, Bool *ismovie)
{
	GeClipMapCallR( IMAGERESULT_NOTEXISTING, InitFromFile )( name, frame, ismovie );
}

void GeClipMap::Destroy()
{
	GeClipMapCall( Destroy )();
}

void GeClipMap::GetDim(LONG *w, LONG *h) const
{
	GeClipMapCall( GetDim )( w, h );
}

LONG GeClipMap::GetBw() const
{
	GeClipMapCallR( NOTOK, GetBw )();
}

LONG GeClipMap::GetBh() const
{
	GeClipMapCallR( NOTOK, GetBh )();
}

BaseBitmap* GeClipMap::GetBitmap()
{
	GeClipMapCallR( NULL, GetBitmap )();
}

void GeClipMap::BeginDraw()
{
	GeClipMapCall( BeginDraw )();
}

void GeClipMap::EndDraw()
{
	GeClipMapCall( EndDraw )();
}

void GeClipMap::SetOffset(LONG off_x, LONG off_y)
{
	GeClipMapCall( SetOffset )( off_x, off_y );
}

void GeClipMap::SetDrawMode(GE_CM_DRAWMODE mode, ULONG par)
{
	GeClipMapCall( SetDrawMode )( mode, par );
}

void GeClipMap::SetColor(LONG r, LONG g, LONG b, LONG a)
{
	GeClipMapCall( SetColor )( r, g, b, a );
}

void GeClipMap::SetPixel(LONG x, LONG y)
{
	GeClipMapCall( SetPixel )( x, y );
}

void GeClipMap::Line(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCall( CMLine )( x1, y1, x2, y2 );
}

void GeClipMap::PolyLine(LONG cnt, GE_POINT2D *points)
{
	GeClipMapCall( PolyLine )( cnt, points );
}

void GeClipMap::FillPolygon(LONG cnt, GE_POINT2D *points)
{
	GeClipMapCall( FillPolygon )( cnt, points );
}

void GeClipMap::Rect(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCall( Rect )( x1, y1, x2, y2 );
}

void GeClipMap::FillRect(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCall( FillRect )( x1, y1, x2, y2 );
}

void GeClipMap::Arc(LONG x1, LONG y1, LONG x2, LONG y2, GE_CM_ARCSEGMENT seg)
{
	GeClipMapCall( Arc )( x1, y1, x2, y2, seg );
}

void GeClipMap::FillArc(LONG x1, LONG y1, LONG x2, LONG y2, GE_CM_ARCSEGMENT seg)
{
	GeClipMapCall( FillArc )( x1, y1, x2, y2, seg );
}

void GeClipMap::Ellipse(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCall( Ellipse )( x1, y1, x2, y2 );
}

void GeClipMap::FillEllipse(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCall( FillEllipse )( x1, y1, x2, y2 );
}

void GeClipMap::SetPixelRGBA(LONG x, LONG y, LONG r, LONG g, LONG b, LONG a)
{
	GeClipMapCall( SetPixelRGBA )( x, y, r, g, b, a );
}

void GeClipMap::GetPixelRGBA(LONG x, LONG y, LONG *r, LONG *g, LONG *b, LONG *a)
{
	GeClipMapCall( GetPixelRGBA )( x, y, r, g, b, a );
}

void GeClipMap::TextAt(LONG x, LONG y, const String &txt)
{
	GeClipMapCall( TextAt )( x, y, txt );
}

LONG GeClipMap::GetTextWidth(const String &txt)
{
	GeClipMapCallR( 0, GetTextWidth )( txt );
}

LONG GeClipMap::GetTextHeight()
{
	GeClipMapCallR( 0, GetTextHeight )();
}

LONG GeClipMap::GetTextAscent()
{
	GeClipMapCallR( 0, GetTextAscent )();
}

void GeClipMap::Blit(LONG dx, LONG dy, const GeClipMap &s_dp, LONG sx1, LONG sy1, LONG sx2, LONG sy2, GE_CM_BLIT rop)
{
	GeClipMapCall( Blit )( dx, dy, *(const iGeClipMap *) &s_dp, sx1, sy1, sx2, sy2, rop );
}

void GeClipMap::SetClipRgn(LONG left, LONG top, LONG right, LONG bottom)
{
	GeClipMapCall( SetClipRgn )( left, top, right, bottom );
}

void GeClipMap::SetClipRgn(GE_POINT2D *points, LONG count, GE_CM_CLIPREGION mode)
{
	GeClipMapCall( SetClipRgn2 )( points, count, mode );
}

Bool GeClipMap::ClipPoint(LONG x, LONG y)
{
	GeClipMapCallR( FALSE, ClipPoint )( x, y );
}

LONG GeClipMap::ClipArea(LONG x1, LONG y1, LONG x2, LONG y2)
{
	GeClipMapCallR( FALSE, ClipArea )( x1, y1, x2, y2 );
}

//----------------------------------------------------------------------------------------
// Set the current font
// Function result:		TRUE if successful
// font_description:	font description (from GeChooseFont(), GetFontDescription() or EnumerateFonts()) or NULL (default font)
// font_size:					requested font size (0.0: the size is read from the font description and if that doesn't have a size a default setting is used)
//----------------------------------------------------------------------------------------
Bool	GeClipMap::SetFont( BaseContainer const *font_description, Real font_size )
{
	GeClipMapCallR( FALSE, SetFont )( font_description, font_size );
}

//----------------------------------------------------------------------------------------
// Get the current font description
// Function result:		font size
// font_description:	used to return the font description
//----------------------------------------------------------------------------------------
Real	GeClipMap::GetFont( BaseContainer *font_description )
{
	GeClipMapCallR( 0.0, GetFont )( font_description );
}

//----------------------------------------------------------------------------------------
// Return a font description for one of the default fonts
// Function result:		TRUE if type is valid
// type:							the requested kind of default font
// font_description:	used to return container with font description (e.g. from GeChooseFont)
//----------------------------------------------------------------------------------------
Bool	GeClipMap::GetDefaultFont( GeFontDefaultType type, BaseContainer *font_description )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, GetDefaultFont));
	if ( lib && lib->GetFontSize )
		return lib->GetDefaultFont( type, font_description );

	return FALSE;
}

//----------------------------------------------------------------------------------------
// Get the font size (in an unspecified unit)
// To get the size in pixels you have to set the font and call GetHeight() and GetAscent()
// Function result:		TRUE if the font size was specified
// font_description:	container with font description (e.g. from GeChooseFont)
// type:							currently GE_FONT_SIZE_INTERNAL only
// size:							used to return the font size
//----------------------------------------------------------------------------------------
Bool	GeClipMap::GetFontSize( BaseContainer const *font_description, GeFontSizeType type, Real *size )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, GetFontSize));
	if ( lib && lib->GetFontSize )
		return lib->GetFontSize( font_description, type, size );

	return FALSE;
}

//----------------------------------------------------------------------------------------
// Set the font size (in an OS specific unit)
// Function result:		TRUE if the font size was specified
// font_description:	container with font description (e.g. from GeChooseFont)
// type:							currently GE_FONT_SIZE_INTERNAL only
// size:							the font size
//----------------------------------------------------------------------------------------
Bool	GeClipMap::SetFontSize( BaseContainer *font_description, GeFontSizeType type, Real size )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, SetFontSize));
	if ( lib && lib->SetFontSize )
		return lib->SetFontSize( font_description, type, size );

	return FALSE;
}

//----------------------------------------------------------------------------------------
// Get a font name from a font description
// Function result:		TRUE if the requested name could be found
// font_description:	container with font description (e.g. from GeChooseFont)
// type:							the requested kind of name (display, family, style or postscript name)
// dst:								used to return the name string
//----------------------------------------------------------------------------------------
Bool	GeClipMap::GetFontName( BaseContainer const *font_description, GeFontNameType type, String *dst )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, GetFontName));
	if ( lib && lib->GetFontName )
		return lib->GetFontName( font_description, type, dst );

	return FALSE;
}

//----------------------------------------------------------------------------------------
// Build a font description
// Function result:		TRUE if the description is valid
// name:							font name
// type:							kind of name (currently only GE_FONT_NAME_POSTSCRIPT is supported)
// dst:								used to return the font description
//----------------------------------------------------------------------------------------
Bool	GeClipMap::GetFontDescription( String const &name, GeFontNameType type, BaseContainer *dst )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, GetFontDescription));
	if ( lib && lib->GetFontDescription )
		lib->GetFontDescription( name, type, dst );

	return FALSE;
}

//----------------------------------------------------------------------------------------
// Enumerate all fonts and return them in a BaseContainer
// If sort_mode is GE_CM_FONTSORT_FLAT for each font a BaseContainer will be inserted that 
// can be used for SetFont() or GetFontName() or as input for GeChooseFont(). The font
// descriptions will be in no particular order
// If sort_mode is GE_CM_FONTSORT_HIERARCHICAL for each family a BaseContainer will be inserted
// that will contain BaseContainers describing the fonts belonging to the family. The families
// are guaranteed to be in alphabetical order
// Function result:		-
// dst:								used to return the font list
// sort_mode:					GE_CM_FONTSORT_HIERARCHICAL: for each font family a BaseContainer is 
//										stored in dst and will hold a description for each font of this family (one 
//										BaseContainer per font). The font families are sorted alphabetically
//										GE_CM_FONTSORT_FLAT: one BaseContainer per font, no order guaranteed
//----------------------------------------------------------------------------------------
void	GeClipMap::EnumerateFonts( BaseContainer *dst, GE_CM_FONTSORT sort_mode )
{
	GeClipMapLib *lib = CheckLib(LIBOFFSET(GeClipMapLib, EnumerateFonts));
	if ( lib && lib->EnumerateFonts )
		lib->EnumerateFonts( dst, sort_mode );
	else
		dst->FlushAll();
}

#endif
