#include <string.h>
#include <stdlib.h>

#include "c4d_memory.h"
#include "operatingsystem.h"
#include "c4d_general.h"

void _GeFree(void **Daten)
{
	if (*Daten)
	{
		C4DOS.Ge->Free(*Daten);
	}
	*Daten=NULL;
}

// only for debug purposes
void _GeCheckAllMemory()
{
	C4DOS.Ge->GeCheckMem((void*)(VULONG)0xdeadbeef);
}

void _GeCheck(void *memptr)
{
	if (memptr)
	{
		C4DOS.Ge->GeCheckMem(memptr);
	}
}

Bool	GeGetAllocSize( void *p, VLONG *out_size )
{
	return C4DOS.Ge->GetAllocSize( p, out_size );
}

// allocator is either NULL (current thread's allocator as used by GeAlloc, gNew etc.) or a pool created via GeAllocPool
Bool	GeGetAllocatorStatistics( BaseContainer &stat, void *allocator )
{
	return C4DOS.Ge->GeGetAllocatorStatistics( stat, allocator );
}

// try to estimate how much physical memory is still unused
VULONG	GeMemGetFreePhysicalMemoryEstimate( void )
{
	return C4DOS.Ge->GeMemGetFreePhysicalMemoryEstimate();
}


#define	STDLIB_MEM_MAGIC		-1
Bool	stdlib_mem_used = FALSE;												// changed to TRUE if static constructors have allocated memory before c4d's memory management was available

Bool	IsAlienMem( void *p )
{
	if ( p )
	{
		if (((VLONG*)p)[-1]==STDLIB_MEM_MAGIC)									// is this a block for the stdlib?
			return TRUE;
	}
	return FALSE;
}

namespace c4d_misc
{

static void *AlienMem(size_t s, Bool clear)
{
	if (s<1) s=1;

	void *p = malloc(s+sizeof(VLONG));
	if (!p) return NULL;

	if (clear) memset(p,0,s+sizeof(VLONG));
	*(VLONG *) p = STDLIB_MEM_MAGIC;
	p = (void *) ((UCHAR *)p + sizeof(VLONG));
	stdlib_mem_used = TRUE;																		// static constructor has allocated memory
	return p;
}

void* GeNewMem(std::size_t s, int line, const char* file, bool clear)	// noexcept
{
	if (t_C4DOS)
	{
		if (clear)
			return C4DOS.Ge->Alloc(s < 1 ? 1 : s, line, file);
		else
			return C4DOS.Ge->AllocNC(s < 1 ? 1 : s, line, file);
	}

	return AlienMem(s, clear);
}

void GeDeleteMem(void* p)	// noexcept
{
	if (p)
	{
		void* temp = p;

		if (stdlib_mem_used)																			// memory allocated by static constructors?
		{
			if (((VLONG*)p)[-1] == STDLIB_MEM_MAGIC)								// is this a block for the stdlib?
			{
				free((void*)((UCHAR*)p - sizeof(VLONG)));
				return;
			}
		}
		C4DOS.Ge->Free(temp);
	}
}

}
