/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef __C4D_VIDEOPOSTPLUGIN_H
#define __C4D_VIDEOPOSTPLUGIN_H

#include "ge_math.h"
#include "c4d_basetime.h"
#include "c4d_videopostdata.h"
#include "c4d_nodeplugin.h"

class GlFrameBuffer;

struct VIDEOPOSTPLUGIN : public NODEPLUGIN
{
	LONG	priority;

	Bool	(VideoPostData::*Draw						)(BaseVideoPost *node, BaseDraw *bd, BaseDrawHelp *bh);
	void	(VideoPostData::*AllocateBuffers)(BaseVideoPost *node, Render *render, BaseDocument *doc);
	VIDEOPOSTINFO (VideoPostData::*GetRenderInfo  )(BaseVideoPost *node);

	RENDERRESULT (VideoPostData::*Execute	)(BaseVideoPost *node, VideoPostStruct *vps);
	void	(VideoPostData::*ExecuteLine		)(BaseVideoPost *node, PixelPost *pp);
	void	(VideoPostData::*ExecutePixel   )(BaseVideoPost *node, PixelPost *pp, LONG x, LONG subx, LONG suby);

	void  (VideoPostData::*CalcVolumetric	)(BaseVideoPost *node, VolumeData *vd);
	void  (VideoPostData::*CalcShadow    	)(BaseVideoPost *node, VolumeData *vd);
	Bool	(VideoPostData::*GlDraw					)(BaseVideoPost *node, BaseDraw *bd, GlFrameBuffer* fbuf, LONG colortex, LONG depthtex, VIDEOPOST_GLDRAW flags);

	Vector (VideoPostData::*ComputeDiffuseGI     )(BaseVideoPost *node, VolumeData *vd, Real receive_strength);
	Vector (VideoPostData::*ComputeDiffuseCaustic)(BaseVideoPost *node, VolumeData *vd, Real sampleradius, LONG accuracy);
	Vector (VideoPostData::*ComputeVolumeCausticEx)(BaseVideoPost *node, const LVector &p);

	void  (VideoPostData::*CreateRay      )(Ray *dst, Real x, Real y);
	VIDEOPOST_GLINFO (VideoPostData::*GetGlInfo)(BaseVideoPost *node, BaseDocument* doc, BaseDraw *bd);
	LONG (VideoPostData::*StereoGetCameraCountEditor)(BaseVideoPost *node, BaseDocument* doc, BaseDraw *bd);
	Bool (VideoPostData::*StereoGetCameraInfo)(BaseVideoPost *node, BaseDocument* doc, BaseDraw *bd, RenderData* rd, LONG index, StereoCameraInfo& info);
	LONG (VideoPostData::*StereoGetSceneHookID)(BaseVideoPost *node);
	Bool (VideoPostData::*StereoMergeImages)(BaseVideoPost *node, BaseBitmap* dest, const BaseBitmap* const* source, LONG cnt, const BaseContainer &settings, COLORSPACETRANSFORMATION transform);
	LONG (VideoPostData::*StereoGetCameraCountRenderer)(BaseVideoPost *node, BaseDocument* doc, RenderData* rd);

	Bool (VideoPostData::*RenderEngineCheck)(BaseVideoPost *node, LONG id);

	Vector (VideoPostData::*ComputeVolumeCaustic )(BaseVideoPost *node, const LVector &p, LONG cpu_index);

	void (VideoPostData::*SetDiffuseWeight)(BaseVideoPost *node, LONG cpu_index, Real weight);

	void (VideoPostData::*CreateExtendedRay)(Ray *dst, Real x, Real y, Real lensx, Real lensy, Real time);
	
	void*	reserved[(32-23)*C4DPL_MEMBERMULTIPLIER-1];
};

#endif
