#ifndef _MATRIX4_H_
#define _MATRIX4_H_

#include "ge_math.h"
#include "vector4.h"

template <class TYPE> class _Matrix4
{
public:
	_Matrix4(_DONTCONSTRUCT v)
	{
	}

	_Matrix4()
	{
		_00 = 1.0f; _01 = 0.0f; _02 = 0.0f; _03 = 0.0f;
		_10 = 0.0f; _11 = 1.0f; _12 = 0.0f; _13 = 0.0f;
		_20 = 0.0f; _21 = 0.0f; _22 = 1.0f; _23 = 0.0f;
		_30 = 0.0f; _31 = 0.0f; _32 = 0.0f; _33 = 1.0f;
	}

	_Matrix4(TYPE m00, TYPE m01, TYPE m02, TYPE m03,
					TYPE m10, TYPE m11, TYPE m12, TYPE m13,
					TYPE m20, TYPE m21, TYPE m22, TYPE m23,
					TYPE m30, TYPE m31, TYPE m32, TYPE m33)
	{
		_00 = m00; _01 = m01; _02 = m02; _03 = m03;
		_10 = m10; _11 = m11; _12 = m12; _13 = m13;
		_20 = m20; _21 = m21; _22 = m22; _23 = m23;
		_30 = m30; _31 = m31; _32 = m32; _33 = m33;
	}

	explicit _Matrix4(TYPE* p)
	{
		LONG i;
		for (i = 0; i < 16; i++)
			m_prArray[i] = p[i];
	}

	_Matrix4(const _Matrix4 &m)
	{
		LONG i;
		for (i = 0; i < 16; i++)
			m_prArray[i] = m.m_prArray[i];
	}

	explicit _Matrix4(const SMatrix &m)
	{
		_00 = (TYPE)m.v1.x; _01 = (TYPE)m.v2.x; _02 = (TYPE)m.v3.x; _03 = (TYPE)m.off.x;
		_10 = (TYPE)m.v1.y; _11 = (TYPE)m.v2.y; _12 = (TYPE)m.v3.y; _13 = (TYPE)m.off.y;
		_20 = (TYPE)m.v1.z; _21 = (TYPE)m.v2.z; _22 = (TYPE)m.v3.z; _23 = (TYPE)m.off.z;
		_30 = (TYPE)0.0;    _31 = (TYPE)0.0;    _32 = (TYPE)0.0;    _33 = (TYPE)1.0;
	}

	explicit _Matrix4(const LMatrix &m)
	{
		_00 = (TYPE)m.v1.x; _01 = (TYPE)m.v2.x; _02 = (TYPE)m.v3.x; _03 = (TYPE)m.off.x;
		_10 = (TYPE)m.v1.y; _11 = (TYPE)m.v2.y; _12 = (TYPE)m.v3.y; _13 = (TYPE)m.off.y;
		_20 = (TYPE)m.v1.z; _21 = (TYPE)m.v2.z; _22 = (TYPE)m.v3.z; _23 = (TYPE)m.off.z;
		_30 = (TYPE)0.0;    _31 = (TYPE)0.0;    _32 = (TYPE)0.0;    _33 = (TYPE)1.0;
	}

	SMatrix GetSMatrix()
	{
		SMatrix m(DC);
		m.v1.x = (SReal)_00; m.v2.x = (SReal)_01; m.v3.x = (SReal)_02; m.off.x = (SReal)_03;
		m.v1.y = (SReal)_10; m.v2.y = (SReal)_11; m.v3.y = (SReal)_12; m.off.y = (SReal)_13;
		m.v1.z = (SReal)_20; m.v2.z = (SReal)_21; m.v3.z = (SReal)_22; m.off.z = (SReal)_23;
		return m;
	}

	LMatrix GetLMatrix()
	{
		LMatrix m(DC);
		m.v1.x = (LReal)_00; m.v2.x = (LReal)_01; m.v3.x = (LReal)_02; m.off.x = (LReal)_03;
		m.v1.y = (LReal)_10; m.v2.y = (LReal)_11; m.v3.y = (LReal)_12; m.off.y = (LReal)_13;
		m.v1.z = (LReal)_20; m.v2.z = (LReal)_21; m.v3.z = (LReal)_22; m.off.z = (LReal)_23;
		return m;
	}

	friend const _Matrix4 operator * (const _Matrix4 &m1, const _Matrix4 &m2)
	{
		LONG i, j, k;
		_Matrix4 r(DC);
		r.SetZero();

		for (i = 0; i < 4; i++)
		{
			for (k = 0; k < 4; k++)
			{
				for (j = 0; j < 4; j++)
				{
					r.m_prElements[i][j] += m1.m_prElements[i][k] * m2.m_prElements[k][j];
				}
			}
		}
		return r;
	}

	friend const LVector4 operator * (const _Matrix4 &m, const LVector4 &v)
	{
		LVector4 r(0.0f);
		LONG i, k;

		for (i = 0; i < 4; i++)
		{
			for (k = 0; k < 4; k++)
			{
				r.m_prElements[i] += m.m_prElements[i][k] * v.m_prElements[k];
			}
		}
		return r;
	}

	friend const SVector4 operator * (const _Matrix4 &m, const SVector4 &v)
	{
		SVector4 r(0.0f);
		LONG i, k;

		for (i = 0; i < 4; i++)
		{
			for (k = 0; k < 4; k++)
			{
				r.m_prElements[i] += m.m_prElements[i][k] * v.m_prElements[k];
			}
		}
		return r;
	}

	friend const SVector operator * (const _Matrix4 &m, const SVector &v)
	{
		SVector4 t(v);
		SVector4 r(0.0f);
		LONG i, k;

		for (i = 0; i < 4; i++)
		{
			for (k = 0; k < 4; k++)
			{
				r.m_prElements[i] += m.m_prElements[i][k] * t.m_prElements[k];
			}
		}
		r.MakeVector3();
		return SVector(r.x, r.y, r.z);
	}

	friend const LVector operator * (const _Matrix4 &m, const LVector &v)
	{
		LVector4 t(v);
		LVector4 r(0.0);
		LONG i, k;

		for (i = 0; i < 4; i++)
		{
			for (k = 0; k < 4; k++)
			{
				r.m_prElements[i] += m.m_prElements[i][k] * t.m_prElements[k];
			}
		}
		r.MakeVector3();
		return LVector(r.x, r.y, r.z);
	}

	_Matrix4 GetTranspose() const
	{
		_Matrix4 r(DC);
		LONG i, j;
		for (i = 0; i < 4; i++)
		{
			for (j = 0; j < 4; j++)
				r.m_prElements[i][j] = m_prElements[j][i];
		}
		return r;
	}

	friend const _Matrix4 operator ! (const _Matrix4 &m)
	{
		_Matrix4 r(DC);

		r._00 =  Det(m._11, m._12, m._13, m._21, m._22, m._23, m._31, m._32, m._33);
		r._01 = -Det(m._01, m._02, m._03, m._21, m._22, m._23, m._31, m._32, m._33);
		r._02 =  Det(m._01, m._02, m._03, m._11, m._12, m._13, m._31, m._32, m._33);
		r._03 = -Det(m._01, m._02, m._03, m._11, m._12, m._13, m._21, m._22, m._23);
		r._10 = -Det(m._10, m._12, m._13, m._20, m._22, m._23, m._30, m._32, m._33);
		r._11 =  Det(m._00, m._02, m._03, m._20, m._22, m._23, m._30, m._32, m._33);
		r._12 = -Det(m._00, m._02, m._03, m._10, m._12, m._13, m._30, m._32, m._33);
		r._13 =  Det(m._00, m._02, m._03, m._10, m._12, m._13, m._20, m._22, m._23);
		r._20 =  Det(m._10, m._11, m._13, m._20, m._21, m._23, m._30, m._31, m._33);
		r._21 = -Det(m._00, m._01, m._03, m._20, m._21, m._23, m._30, m._31, m._33);
		r._22 =  Det(m._00, m._01, m._03, m._10, m._11, m._13, m._30, m._31, m._33);
		r._23 = -Det(m._00, m._01, m._03, m._10, m._11, m._13, m._20, m._21, m._23);
		r._30 = -Det(m._10, m._11, m._12, m._20, m._21, m._22, m._30, m._31, m._32);
		r._31 =  Det(m._00, m._01, m._02, m._20, m._21, m._22, m._30, m._31, m._32);
		r._32 = -Det(m._00, m._01, m._02, m._10, m._11, m._12, m._30, m._31, m._32);
		r._33 =  Det(m._00, m._01, m._02, m._10, m._11, m._12, m._20, m._21, m._22);
		TYPE det = m._00 * r._00 + m._10 * r._01 + m._20 * r._02 + m._30 * r._03;
		if (det == 0)
		{
			r = _Matrix4();
			return r;
		}
    det = 1.0f / det;
		r = det * r;
		return r;
	}

	friend const _Matrix4 operator * (const TYPE s, const _Matrix4 &m)
	{
		_Matrix4 r(DC);
		LONG i;
		for (i = 0; i < 16; i++)
			r.m_prArray[i] = m.m_prArray[i] * s;
		return r;
	}

	friend Bool operator == (const _Matrix4& a, const _Matrix4& b)
	{
		for (LONG l = 0; l < 16; l++)
		{
			if (a.m_prArray[l] != b.m_prArray[l]) return FALSE;
		}
		return TRUE;
	}

	friend Bool operator != (const _Matrix4& a, const _Matrix4& b)
	{
		for (LONG l = 0; l < 16; l++)
		{
			if (a.m_prArray[l] != b.m_prArray[l]) return TRUE;
		}
		return FALSE;
	}

	void SetZero()
	{
		_00 = _01 = _02 = _03 = 0.0f;
		_10 = _11 = _12 = _13 = 0.0f;
		_20 = _21 = _22 = _23 = 0.0f;
		_30 = _31 = _32 = _33 = 0.0f;
	}

	void SetOffset(const Vector& v)
	{
		_03 = v.x;
		_13 = v.y;
		_23 = v.z;
	}

	void Scale(TYPE r)
	{
		_33 /= r;
	}

	void ScaleDirections(TYPE r)
	{
		_00 *= r; _01 *= r; _02 *= r;
		_10 *= r; _11 *= r; _12 *= r;
		_20 *= r; _21 *= r; _22 *= r;
	}

	Bool IsInvertable()
	{
		_Matrix4 r(DC);

		r._00 =  Det(_11, _12, _13, _21, _22, _23, _31, _32, _33);
		r._01 = -Det(_01, _02, _03, _21, _22, _23, _31, _32, _33);
		r._02 =  Det(_01, _02, _03, _11, _12, _13, _31, _32, _33);
		r._03 = -Det(_01, _02, _03, _11, _12, _13, _21, _22, _23);
		r._10 = -Det(_10, _12, _13, _20, _22, _23, _30, _32, _33);
		r._11 =  Det(_00, _02, _03, _20, _22, _23, _30, _32, _33);
		r._12 = -Det(_00, _02, _03, _10, _12, _13, _30, _32, _33);
		r._13 =  Det(_00, _02, _03, _10, _12, _13, _20, _22, _23);
		r._20 =  Det(_10, _11, _13, _20, _21, _23, _30, _31, _33);
		r._21 = -Det(_00, _01, _03, _20, _21, _23, _30, _31, _33);
		r._22 =  Det(_00, _01, _03, _10, _11, _13, _30, _31, _33);
		r._23 = -Det(_00, _01, _03, _10, _11, _13, _20, _21, _23);
		r._30 = -Det(_10, _11, _12, _20, _21, _22, _30, _31, _32);
		r._31 =  Det(_00, _01, _02, _20, _21, _22, _30, _31, _32);
		r._32 = -Det(_00, _01, _02, _10, _11, _12, _30, _31, _32);
		r._33 =  Det(_00, _01, _02, _10, _11, _12, _20, _21, _22);
		TYPE det = _00 * r._00 + _10 * r._01 + _20 * r._02 + _30 * r._03;
		if (det == 0.0)
		{
			return FALSE;
		}
		return TRUE;
	}

	union
	{
		struct { TYPE _00, _01, _02, _03, 
									 _10, _11, _12, _13,
									 _20, _21, _22, _23,
									 _30, _31, _32, _33; };
		TYPE m_prElements[4][4];
		TYPE m_prArray[16];
	};

private:
	static TYPE Det(TYPE a1, TYPE a2, TYPE a3, TYPE b1, TYPE b2, TYPE b3, TYPE c1, TYPE c2, TYPE c3)
	{
		return a1 * (b2 * c3 - b3 * c2) - b1 * (a2 * c3 - a3 * c2) + c1 * (a2 * b3 - a3 * b2);
	}
};

typedef _Matrix4<SReal> SMatrix4;
typedef _Matrix4<LReal> LMatrix4;

#ifdef __SINGLEPRECISION
	#define Matrix4 SMatrix4
	#define RMtoSM4(x)	(x)
	#define SMtoRM4(x)	(x)
	inline LMatrix4 RMtoLM4(const Matrix4 m)
	{
		return LMatrix4((LReal)m._00, (LReal)m._01, (LReal)m._02, (LReal)m._03,
										(LReal)m._10, (LReal)m._11, (LReal)m._12, (LReal)m._13,
										(LReal)m._20, (LReal)m._21, (LReal)m._22, (LReal)m._23,
										(LReal)m._30, (LReal)m._31, (LReal)m._32, (LReal)m._33);
	}
#else
	#define Matrix4 LMatrix4
	inline SMatrix4 RMtoSM4(const Matrix4 m)
	{
		return SMatrix4((SReal)m._00, (SReal)m._01, (SReal)m._02, (SReal)m._03,
										(SReal)m._10, (SReal)m._11, (SReal)m._12, (SReal)m._13,
										(SReal)m._20, (SReal)m._21, (SReal)m._22, (SReal)m._23,
										(SReal)m._30, (SReal)m._31, (SReal)m._32, (SReal)m._33);
	}
	inline Matrix4 SMtoRM4(const SMatrix4& m)
	{
		return Matrix4((Real)m._00, (Real)m._01, (Real)m._02, (Real)m._03,
										(Real)m._10, (Real)m._11, (Real)m._12, (Real)m._13,
										(Real)m._20, (Real)m._21, (Real)m._22, (Real)m._23,
										(Real)m._30, (Real)m._31, (Real)m._32, (Real)m._33);
	}
	#define RMtoLM4(m) (m)
#endif

#endif // _MATRIX4_H_
